//
// Stepper control
//

#include <P16f877.h>
#include <datalib.h>
#include <displays.h>

//
// Define software version string
//

const unsigned char VersionString[]="V2.03  20-Feb-07";

#include "step2-vars.h"

// FUNCTION PROFORMAS

// Clear LCD display

void clearLCD();

// Output N-digit number to LCD

void LCDNum(long number, char digits, char point);

// Edit a number on the screen

long LCDEditNum(long number, char digits,
		int column, char *lineOne, char point, long max, long min);

// Print step number as degrees on line 2

void LCDstepNumber();

// Set parameters to default values

void DefaultParams();

// Get and convert current key stroke

unsigned char getKey();

// Wait for a stable change of key press

void keyChange();

// Restore current settings

void restoreSettings(char parameterSet);

// Save current settings

void saveSettings(char parameterSet);

// Ramp up/down
void rampud(long steps);

// Ramp continuous
void rampcont();

// Do defined number of steps
void dosteps(long steps, char rampRate);

//Time delay
void waitTime(unsigned int time);

// Select the mode
void selectMode(char *modeName, char *line2);

//Set Current procedure

void setCurrent(int percent);

//Move Degrees procedure

void moveDegrees(long degs);

// Convert Speed procedure

unsigned int convertSpeed(long speed);

// Set Speeds procedure

char setSpeeds();

//
// MAIN MODE FUNCTIONS
//

// Setup mode
void setup();

//Jog mode
void jog();

//Division mode
void division();

//Degrees mode
void degrees();

//Continuous mode
void continuous();

//Slave (4th Axis) mode
void slave();

//Function proforma for interrupt handler

//void Interrupt();

//
//MAIN
//

void main ()
{

//Do initialisation functions

 ADCON1=0x7;			//Set no analogue inputs
 TRISA=0x0;			//Set all bits of Port A to outputs
 OPTION_REG=0;			//Enable pull-ups on B, Timer 0 prescale =2,
 				//Timer 0 internal clock source
 				
// INTCON |=(1<<T0IE)|(1<<GIE)|(1<<PEIE);
 				//Enable timer 0 interrupts
// #asm
// bcf INTCON, T0IF		; Clear the timer interrupt flag
// #asmend

 TRISB=0xF8;			//Set bitS 0 to 2 of port B to output,
 				//the rest to input
 TRISE=0x7;			//Set Port E bits 0-2 to input
 PB.B0=1;			//Disable motor, direction low, clock high
 PB.B1=0;
 PB.B2=0;
 TRISD=0;			//All bits are outputs (Display I/O bits)
 PD.B0=0;			//Signal "Stopped"
  
 PR2=255;			//Set the PWM period (Timer 2 period reg)
 CCPR1L=0;			//Set the duty cycle to zero for the moment
 PORTC=0;			//Clear the output buffers
 TRISC=0xFB;			//Bit 2 of C is output, rest are input
 T2CON=0x04;			//Timer 2 on, Prescale=1, Postscale=1
  
// Initialize the LCD with startup message

 LCD(-2);			//2-line display, 8-bit, 5x7 format
 clearLCD();			//Initialize & clear screen
 LCDString (" DivisionMaster");//Hello to the world
 LCDPrintAt(0,1);		//2nd Line
 LCDString (VersionString);	//Display software version
 
// Load the current parameter set

 currentParamSet=ReadEEData(0);	//Read last used param set from EEProm

// Check for first time use and initialize parameter sets if so

 if ((currentParamSet<0)||(currentParamSet>=maxParamSet))
 { 				//If this is the first time...
  DefaultParams();		//Set to default settings
  for (currentParamSet=(maxParamSet); currentParamSet; currentParamSet--)
   saveSettings(currentParamSet-1);
   				//Save current settings in all parameter sets
 }

// Make sure all parameters are correctly loaded
 
 restoreSettings(currentParamSet);
 				//Restore current settings from NV Ram
 				
// Check for slave mode

 newKey=oldKey=getKey();	//Initialize key press variables
 if(newKey==Abort)		//Abort held down?
  slave();			//Slave mode if so...
   				//There is no exit from this mode 
   				//other than powering down
   				
 for (delay=15; delay; delay--)
  waitTime(32000);		//Wait a bit to show off the startup screen

 
//  Check whether the user wants to switch to a different parameter set
 
 currentParamSet=LCDEditNum(currentParamSet, 1, 15, "Parameter set=", 9, maxParamSet-1, 0);
 				//Edit the current parameter set (use rate as temp var)
 				
// Make sure all parameters are correctly loaded
 
 restoreSettings(currentParamSet);
 				//Restore current settings from NV Ram

// MAIN CONTROL LOOP - Select major mode & execute that mode's behaviour
 
 while(1)			//Loop forever
 {
modeJog:

  clearLCD();			//Clear the screen
  selectMode("JOG mode", EnterString);
  if (newKey==Ok)
  {
   jog();			//See if mode is "Jog"
  }
  if (newKey==Up)
   goto modeSetup;
 
modeDivision:

  clearLCD();			//Clear the screen
  selectMode("DIVISION mode", EnterString);
  if (newKey==Ok)
  {
   division();			//See if mode is "Division"
  }
  if (newKey==Up)
   goto modeJog;

modeDegrees:

  clearLCD();			//Clear the screen
  selectMode("DEGREES mode", EnterString);
  if (newKey==Ok)
  {
   degrees();			//See if mode is "Degrees"
  }
  if (newKey==Up)
   goto modeDivision;

modeContinuous:

  clearLCD();			//Clear the screen
  selectMode("CONTINUOUS mode", EnterString);
  if (newKey==Ok)
  {
   continuous();		//See if mode is "Continuous"
  }
  if (newKey==Up)
   goto modeDegrees;

modeSetup:

  clearLCD();			//Clear the screen
  LCDPrintAt(15,0);		//Position for set number
  LCDNum(currentParamSet, 1, 9);//Print set number to screen
  selectMode("SETUP paramSet", EnterString);
  if (newKey==Ok)
   setup();			//See if mode is "Setup"
  if (newKey==Up)
   goto modeContinuous;

  };
};

//
// rampud function - ramps up/down using the requested values for
// start & end speed, stepping a requested number of steps in the
// requested direction defined by direction, at start/end speeds
// defined by slow and fast.
// direction = 1 for +ve move, -1 for -ve move
// If dir = -1, backlash compensation is invoked
//

void rampud(long steps)
{
 long plateau;			//plateau size
 long ramp;			//ramp size
 if (direction<0)
  steps=steps+Params.Lash;	//Add in backlash steps if -ve move
  
 ramp=(slow-fast)/rate;		//Find out max steps for ramp
 ramp=ramp*8;			//Adjust for divide-by-8 in doSteps
 
 if ((ramp<<1)>(steps))		//Limit ramp size...
  ramp=(steps>>1);		//to half the requested steps
 plateau=steps-(ramp<<1);	//number of steps for plateau - can be zero
 delay=slow;			//set initial speed
 
 PD.B0=1;			//Signal "Running"
 dosteps(ramp, rate);		//do up ramp

 if (newKey==Abort)
  goto abortRamp;		//Bomb out if abort pressed

 if (plateau) 
 {
  dosteps(plateau, 0);		//...and do the plateau

  if (newKey==Abort)
   goto abortRamp;		//Bomb out if abort pressed
 }
 dosteps(ramp, -rate);		//do down ramp

 if (newKey==Abort)
  goto abortRamp;		//Bomb out if abort pressed

 if (direction<0)		//If -ve direction...
 {
  waitTime(1000);		//Give the motor a chance
  direction=1;			//Force a +ve move
  dosteps(Params.Lash, 0);	//Do the backlash steps in +ve direction
 }
abortRamp: 
 PD.B0=0;			//Signal "Stopped"

};
 


//
// Waittime function - waits for a specified delay period.
//

void waitTime(unsigned int time)
{
 while(time--)	;		//Repeat loop time times
};

//
// LCD Clear function
//

void clearLCD()
{
 LCD(0x108);			//Display off
 LCD(0x101);			//Clear display, cursor home
 LCD(0x10f);			//Display on, cursor on, flash cursor
 LCD(0x180);			//Print position 0,0
};

//
// LCDNum function - prints N digit number to LCD display (backwards)
// Includes decimal point if the point parameter is within range of digits
//

void LCDNum(long number, char digits, char point)
{
 LCD(0x104);			//Force reverse cursor
 for(digits; digits; digits--)
 {
  if (digits==point)
   LCD('.');
  else
  {
   LCD((number%10)+'0');	//Print Nth digit 
   number=number/10;		//lose Nth digit
  }
 }
 LCD(0x106);			//Force normal cursor
};

//
// DefaultParams function
//

void DefaultParams()
{
 Params.Ratio=90;		//Worm ratio (max 99)
 Params.CW=1;			//1 if +ve step is CW; 0 if CCW
 Params.Divs=60;		//Default to 60 divisions
 Params.Degrees=30000;		//Default to 30 degrees
 Params.JogSlow=200;		//Jog slow speed
 Params.JogFast=400;		//Jog fast speed
 Params.JogRate=8;		//Jog Ramp Rate
 Params.Slow=350;		//Slow speed
 Params.Fast=700;		//Fast speed
 Params.Rate=8;			//Ramp rate
 Params.Lash=0;			//Rotary backlash (steps)
 Params.Current=15;		//Current is 1.1 amps
 Params.Idle=50;		//Idle current is 50%
};

//
// function - converts raw key press to 0-9 plus function keys
//

unsigned char getKey()
{
 unsigned char x;		//Local var
 
 if (!PE.B2)
  x=Abort;			//External Abort
 else if (!PE.B1)
  x=Right;			//External Right
 else if (!PE.B0)
  x=Left;			//External Left
 else				//Look for keypad strokes
 {
  x=KeyScan();			//Get key stroke
  if (x!=NoKey)
   x=keys[x];			//Translate key stroke
 }
 return x;			//Exit
};

//
// restoreSettings function
// restores last set of parameter settings
//

void restoreSettings(char parameterSet)
{
 char i;			//Loop control
 CharPtr=(&Params.CW)+parSize-1;//Point at the parameter structure
 
 for (i=parSize; i; i--)	//Loop for all chars of the structure
 {
  *CharPtr=ReadEEData(i+(parameterSet*parSize));
  				//Read a character from EEProm to structure
  CharPtr--;			//Point at next slot
 }
 stepsPerRev=Params.Ratio;	//***Before exiting...
 stepsPerRev*=400;		//***Work out how many steps makes 360 degrees
 WriteEEData(0, parameterSet);	//Save the current parameter set number
};

//
// saveSettings function
// saves current set of parameter settings
//

void saveSettings(char parameterSet)
{
 char i;			//Loop control
 WriteEEData(0, parameterSet);
 				//Save the current parameter set number
 CharPtr=(&Params.CW)+parSize-1;//Point at the end of the parameter structure
 
 for (i=parSize; i; i--)	//Loop for all chars of the structure
 {
  WriteEEData(i+(parameterSet*parSize), *CharPtr);
  				//Write a character from structure to EEProm
  CharPtr--;			//Point at next slot
 }
 stepsPerRev=Params.Ratio;		//***Before exiting...
 stepsPerRev*=400;			//***Work out how many steps makes 360 degrees
};

//
// keyChange function
// waits for a stable change of keypress
// including pressed to not-pressed
// Sets oldKey to value of newKey on entry
// Sets newKey to new key value
//

void keyChange()
{
 oldKey=newKey;		// Save old keystroke
 do
 {
  newKey=getKey();	// Get current keystroke
  waitTime(5000);	// Delay a bit
 }
 while((newKey!=getKey()) || (newKey==oldKey));
};

//
// Display "select" string
// Exit if Up, Down or Ok keys pressed
// sets newKey
//

void selectMode(char *modeName, char *line2)
{
 LCDPrintAt(0,0);		//Point to first line
 LCDString(modeName);		//Set the mode name
 LCDPrintAt(0,1);		//Point to second line
 LCDString(line2);		//Set second line text
 do
 {
  keyChange();			//Wait for a change of key
 }
 while ((newKey!=Up) && (newKey!=Down) && (newKey!=Ok));
};

//
//LCDEditNum function - displays an integer (max 10 digits)
//at a defined position on the first row
//and allows the user to "over-write" the value with a new value.
//Returns the new value.
//Supports fixed decimal point - must be WITHIN the number field (not at
//one end or the other). Number of digits includes dec point if enabled.
//Dec point position well outside range of "digits" disables dec point display.
//Parameter max defines an upper limit for the number: this value will be
//forced and re-displayed if operator tries to enter number that is too big
//or too small.
//

long LCDEditNum(long number, char digits,
		int column, char *lineOne, char point, long max, long min)
{
 char c[10];			//Not more than 10 digits
 char i;			//Loop control var

 clearLCD();
 LCDPrintAt(0,0);
 LCDString(lineOne);		//Print frirst line text
 LCDPrintAt(0,1);		//Second line
 LCDString(OkString);		//Second line string
toobig:
 LCD(0x180+column);		//Set print position for number
 for (i=digits; i; i--)		//Unpack the number
 {
  if (i==point)
   c[i-1]='.';			//Set the point if it is there
  else
  {
   c[i-1]=((number%10)+'0');	//Pack bottom digit - in ascii
   number=number/10;		//lose bottom digit
  }
 }				//next digit
 c[digits]=0;			//terminate the string
 LCDString(c);			//Print the string
 
 i=0;				//Initialize index value
 do
 {
  LCD(0x180+column+i);		//point display at current char
  keyChange();			//get a change of key
  if(newKey<10)			//digit pressed?
  {
   c[i]=newKey+'0';		//store ascii digit in current position
   LCD(c[i]);			//Print digit
   ++i;				//increment index
   if (c[i]=='.') ++i;		//don't overwrite the decimal point
  }
  if (newKey==Right)
  { 
   ++i;				//cursor right
   if (c[i]=='.') ++i;		//but don't overwrite the decimal point
  }
   
  if(newKey==Left)
  {
   --i;				//cursor left
   if (c[i]=='.') --i;		//but don't overwrite the decimal point
  }
  if(i<0) i+=digits;		//Don't fall off at the left...
  i=i%digits;			//...or at the right
 }
 while((newKey != Ok) && (newKey != Up) && (newKey != Down));
 				//Continue till ok or up/down is pressed
 				
 number=0;			//Clear the number
 for (i=0; i<digits; i++)	//Re-pack the number
 {
  if (c[i]=='.')
   i++;				//Skip the point if it is there
  number=(number*10)+(c[i]-'0');//Pack a digit
 }				//next digit
 if (number>max)
 {
  number=max;
  goto toobig;			//Impose upper limit
 }
 if (number<min)
 {
  number=min;
  goto toobig;			//Impose lower limit
 }
 return number;			//Return the result
};

//
// Jog function
//

void jog()
{
 long jdegs;
 
 LCDPrintAt(0,1);		//2nd Line
 LCDString(ExitString);		//Display exit command
 stepNumber=position=0;		//Initialize the step number and position
 slow=convertSpeed(Params.JogSlow);
 fast=convertSpeed(Params.JogFast);
 				//Set the jogging speeds
 rate=Params.JogRate;
 do
 {
  LCDPrintAt(15,0);
  LCDNum(position, 7, 4);	//Print target position
  LCDPrintAt(16,0);		//Lose the cursor
  direction=1;			//Assume +ve move
  
  keyChange();			//Wait for a change of key
  
  jdegs=10;			//Hundredth of a degree?
  if (newKey==Left) 
  {
   goto dojogMinus;		//Do the jog backwards
  }
  if (newKey==Right)
  {
   goto dojog;			//Do the jog forwards
  }
  
  jdegs=100;			//Tenth of a degree??
  if (newKey==1)
  {
   goto dojogMinus;		//Do the jog backwards
  }
  if (newKey==3)
  {
   goto dojog;			//Do the jog forwards
  }
  jdegs=1000;			//One degree??
  if (newKey==4)
  {
   goto dojogMinus;		//Do the jog backwards
  }
  if (newKey==6)
  {
   goto dojog;			//Do the jog forwards
  }
  jdegs=10000;			//Ten degrees
  if (newKey==7)
  {
   goto dojogMinus;		//Do the jog backwards
  }
  if (newKey==9)
  {
   goto dojog;			//Do the jog forwards
  }
  continue;			//Loop if not a valid jog command
dojogMinus:			//Here if -ve move
  direction=-1;  
dojog:				//Here if +ve move

  setCurrent(100);		//Set to full current
  moveDegrees(jdegs);
  				//Make the move
  waitTime(10000);		//Wait for a lot - allow motor to stop moving
  setCurrent(Params.Idle);	//...before setting the current to low
  waitTime(10000);		//Wait for a lot - allow motor to stop moving
  setCurrent(5);		//...before setting the current to very low
  PB.B2=0;			//...and disabling the motor
  }
 while ((newKey!=Down) && (newKey!=Up));
 				//Continue jogging until exit key pressed
};


//
// Division mode function
//

void division()

{
 stepNumber=0;			//Set current step & global position to zero
 
 while (1)			//Continue in mode until "up" pressed
 {
  Params.Divs=LCDEditNum(Params.Divs, 4, 12, "Divisions=", 9, 9999, 1);
 				//Edit the number
   
  if (newKey==Up) break;	//Get out if end of operation
  
  position=(stepNumber*Params.Divs)/stepsPerRev;
  				//Calculate the nearest division
  				//to current angular position
  LCDPrintAt(0,0);
  LCDString("Div 0000 Of");	//Set up for the step sequence
  LCDPrintAt(0,1);
  LCDString(ExitString);	//Display the exit command
  do
  {
   LCDPrintAt(7,0);
   LCDNum(position,4, 99);	//Print current step number
   LCDPrintAt(15,0);
   LCDNum(Params.Divs,4, 99);	//Print current max step
   LCDPrintAt(16,0);		//Lose the cursor
   
   if ((!setSpeeds())&&(newKey!=Down))
			   	//Get fast/slow speeds & see if a move needed
    continue;			//If not needed...Don't do a move
   if (newKey==Down)		//Edit the current division number?
   {
    increment=LCDEditNum(position, 4, 6, "Go to", 9, (Params.Divs-1), 0);
				//Edit the number; use increment as local store
    LCDPrintAt(0,0);
    LCDString("Div 0000 Of");	//Repaint the screen
    LCDPrintAt(7,0);
    LCDNum(increment,4, 99);	//Print current step number
    LCDPrintAt(15,0);
    LCDNum(Params.Divs,4, 99);	//Print current max step
    LCDPrintAt(0,1);
    LCDString(ExecuteString);	//Display the execute command
    position=increment;		//Set new position if not
   if (!setSpeeds())		//Get the directional information
     continue;			//...or not as the case may be
   }
   else  position+=direction;	//Bump the current position up or down
   
   if (position<0)
    position+=Params.Divs;	//Correct underflow in position
   position=(position%Params.Divs);
   				//Correct overflow in position
   increment=(((position*stepsPerRev)+(Params.Divs>>1))/Params.Divs);
   				//Calculate number of steps for desired position
   if (direction==1)		//+ve move
    increment-=stepNumber;	//Calculate size of move
   else				//-ve move
    increment=stepNumber-increment;
   				
   if (increment<0)
    increment+=stepsPerRev;	//Correct underflow in increment
   
   if (!increment)
    increment=stepsPerRev;	//Special case when Params.Divs=1
   
   setCurrent(100);		//Set the current to high
 
   rampud(increment);
  				//Do the move
   
   waitTime(10000);		//Let motor settle
   setCurrent(Params.Idle);	//Set current to half 
   
   if (stepNumber<0)
    stepNumber+=stepsPerRev;	//Correct underflow
   stepNumber=(stepNumber%stepsPerRev);
   				//Correct overflow
   LCDstepNumber();		//Print actual position in degrees
  }
  while ((newKey!=Down)&&(newKey!=Up));
 				//Stay in mode till kicked out
 }
 waitTime(10000);		//Wait for a lot - allow motor to stop moving
 PB.B2=0;			//...before disabling the motor
};

//
// Setup mode procedure
//

void setup()
{
 char parameterSet;			//Local storage

 parameterSet=LCDEditNum(currentParamSet, 1, 15, "Parameter set=", 9, maxParamSet-1, 0);
 					//Edit the current parameter set
 if (parameterSet != currentParamSet)	//Change of parameters?
 {
  currentParamSet=parameterSet;
  restoreSettings(currentParamSet);
 }
setFactory:
 selectMode("Factory setting?", EnterString);	//See if mode is "Factory"
 if (newKey==Up) goto saveExit;
 if (newKey==Ok) DefaultParams();
 
restore:
 selectMode("Restore setting?", EnterString);	//See if mode is "Restore"
 if (newKey==Up) goto setFactory;
 if (newKey==Ok) restoreSettings(currentParamSet);
 
currentSet:
 Params.Current=LCDEditNum(Params.Current, 3, 13, "Max amps=", 2, 20, 1);
 					//Edit the max motor current (max 2A)
 if (newKey==Up) goto restore;

setIdle:
 Params.Idle=LCDEditNum(Params.Idle, 3, 12, "Idle amps=     %", 9, 100, 0);
 					//Edit the idle motor current
 if (newKey==Up) goto currentSet;

editRatio:
 Params.Ratio=LCDEditNum(Params.Ratio, 4, 12, "Worm ratio=", 9, 5000, 1);
 					//Edit the ratio
 if (newKey==Up) goto setIdle;

editCW:
 clearLCD();				//Clear the screen
 LCDString("+ve move =");		//See if we want to change CW/CCW
 LCDPrintAt(0,1);			//2nd line
 LCDString(OkString);

showChange:
 LCDPrintAt(13,0);
 if (Params.CW)
  LCDString("CW ");			//Print current direction
 else
  LCDString("CCW");
 
cwGetKey:
 keyChange();				//Get a key stroke
 if (newKey==Down) goto editDivs;	//Next param
 if (newKey==Up) goto editRatio;
 if ((newKey!=Right) && (newKey!=Left)) goto cwGetKey;
 					//get another key
 if (Params.CW)
  Params.CW=0;
 else
  Params.CW=1;				//Flip the state of CW
 goto showChange;			//Show the change

editDivs:
 Params.Divs=LCDEditNum(Params.Divs, 4, 12, "Divisions=", 9, 9999, 1);
 					//Edit # of divisions
 if (newKey==Up) goto editCW;

editDegrees:
 Params.Degrees=LCDEditNum(Params.Degrees, 7, 9, "Degrees=", 4, 360000, 1);
 					//Edit # of degrees
 if (newKey==Up) goto editDivs;

editJogSlow:
 Params.JogSlow=LCDEditNum(Params.JogSlow, 4, 12, "Slow Min=", 9, 9999, 3);
 					//Edit slow jog speed
 if(Params.JogSlow>Params.JogFast) Params.JogFast=Params.JogSlow;
 					//Force proper ordering
 if (newKey==Up) goto editDegrees;

editJogFast:
 Params.JogFast=LCDEditNum(Params.JogFast, 4, 12, "Slow Max=", 9, 9999, 3);
 					//Edit fast jog speed
 if(Params.JogSlow>Params.JogFast) Params.JogSlow=Params.JogFast;
 					//Force proper ordering
 if (newKey==Up) goto editJogSlow;
 
editJogRate:
 Params.JogRate=LCDEditNum(Params.JogRate, 2, 14, "Slow Ramp=", 9, 99, 1);
 					//Edit jog ramp rate
 if (newKey==Up) goto editJogFast;

editSlow:
 Params.Slow=LCDEditNum(Params.Slow, 4, 12, "Fast Min=", 9, 9999, 3);
 					//Edit slow speed
 if(Params.Slow>Params.Fast) Params.Fast=Params.Slow;
 					//Force proper ordering
 if (newKey==Up) goto editJogRate;
 
editFast:
 Params.Fast=LCDEditNum(Params.Fast, 4, 12, "Fast Max=", 9, 9999, 3);
 					//Edit fast speed
 if(Params.Slow>Params.Fast) Params.Slow=Params.Fast;
 					//Force proper ordering
 if (newKey==Up) goto editSlow;

editRate:
 Params.Rate=LCDEditNum(Params.Rate, 2, 14, "Fast Ramp=", 9, 99, 1);
 					//Edit fast ramp rate
 if (newKey==Up) goto editFast;

editLash:
 Params.Lash=LCDEditNum(Params.Lash, 3, 13, "Backlash=", 9, 999, 0);
 					//Edit backlash
 if (newKey==Up) goto editRate;

saveExit:

 selectMode("Save and exit?  ", EnterString);	//See if mode is "Save"
 if (newKey==Up) goto editLash;
 if (newKey==Down) goto setFactory;
 saveSettings(currentParamSet);
 
};

//
// Degrees mode function
//

void degrees()
{
 stepNumber=position=0;			//Set current step &
 					//desired position to zero
 
 while (1)				//Continue in mode until "up" pressed
 {
  Params.Degrees=LCDEditNum(Params.Degrees, 7, 9, "Degrees=", 4, 360000, 1);
 					//Edit the number
  if (newKey==Up) break;		//Get out if end of operation

  clearLCD();
  LCDPrintAt(0,1);
  LCDString(ExitString);		//Set up the exit string
  do
  {
   LCDPrintAt(6,0);
   LCDNum(position, 7, 4);		//Print target position in degrees
   LCDPrintAt(15,0);
   LCDNum(Params.Degrees,7, 4);		//Print current increment in degrees
   LCDPrintAt(16,0);			//Lose the cursor

   if ((!setSpeeds())&&(newKey!=Down))
			   	//Get fast/slow speeds & see if a move needed
    continue;			//If not needed...Don't do a move
   if (newKey==Down)		//Edit the current division number?
   {
    increment=LCDEditNum(position, 7, 6, "Go to", 4, 360000, 0);
				//Edit the target angle; use increment as local store
     
    clearLCD();
    LCDPrintAt(6,0);
    LCDNum(increment, 7, 4);	//Print target position in degrees
    LCDPrintAt(15,0);
    LCDNum(Params.Degrees,7, 4);//Print current increment in degrees
    LCDPrintAt(0,1);
    LCDString(ExecuteString);	//Display the execute command
    if (!setSpeeds())		//Get the directional information
     continue;			//...or not as the case may be
     
    if (direction==1)		//If going clockwise...
     increment-=position;	//Get difference
    else
     increment=position-increment;
    if (increment<0)
     increment +=360000;	//Adjust if necessary
    setCurrent(100);		//High motor current
    moveDegrees(increment);	//Do absolute increment

   }
   else
   {   
    setCurrent(100);			//High motor current
    moveDegrees(Params.Degrees);	//Do normal increment
   }
   					//Make the move
   waitTime(10000);			//Let motor settle
   setCurrent(Params.Idle);		//Set current to half 

  }
  while ((newKey!=Down)&&(newKey!=Up));
 					//Stay in mode till kicked out
 }
 waitTime(10000);			//Wait for a lot - allow motor to stop moving
 PB.B2=0;				//...before disabling the motor
};

//
// LCDstepNumber()
// Print step number as degrees on line 2
//

void LCDstepNumber()
{
 long x;				//Local scratch
 x=((stepNumber*900)+(Params.Ratio>>1))/Params.Ratio;
 					//x=stepNumber as degrees
 LCDPrintAt(9,1);
 LCDNum(x, 7, 4);			//Print stepNumber as degrees
 LCDPrintAt(10,1);
 LCDString("deg");			//Say it is degrees
 LCDPrintAt(16,0);			//Lose the cursor
};

// 
// dosteps function - does a number of steps in the requested direction.
// (global var) direction = 1 for +ve move, -1 for -ve move
// uses the CW param to determine whether +ve move is CW or CCW
// enables motor & exits with motor enabled
// Port B Bit 0 is Step
// Port B Bit 1 is Dir
// Port B Bit 2 is Enable
// Port E Bit 2 is Abort
//

void dosteps(long steps, char rampRate)
{
 int x;				//Local variables
 long steploc;
 char dirloc;
 steploc=steps;			//Local copies of params
 dirloc=direction;
 if (direction<0) dirloc=0;		//Set local dir to 0 or 1
 PB.B1=dirloc^Params.CW;	//Set motor direction bit to effective direction
 PB.B2=1;			//Enable motor
 PA.B1=0;			//Pre-set the keyboard rows
 PA.B2=1;
 PA.B3=0;
 PA.B5=0;
 if ((!PE.B2) || (!PB.B6))
  newKey=Abort;			//External Abort
 dirloc=0;
 while((newKey!=Abort) && steps)
 				//while abort not pressed && steps to do
 {
  x=delay;			//Local copy of delay value for delay loop
  PB.B0=0;			//force bit 0 of port A low
  while(x--) ;			//Wait for current delay period
  PB.B0=1;			//force bit 0 of port A high
  steps --;			//decrement step count
  if (!(steps&7))		//Every eighth iteration...
   delay-=rampRate;		//ramp the rate up or down
  else
   delay-=dirloc;		//otherwise, leave it be (dirloc = 0 here)
   				//(but maintain the code length!)
  if ((!PE.B2) || (!PB.B6))
   newKey=Abort;		//External Abort
 }
 PA.B2=0;			//Reset the keyboard row
 stepNumber+=((steploc-steps)*direction);
 				//adjust global position to account for steps done
};

//
//Move Degrees procedure
//
//Makes a move of a specified number of thousandths of a degree
//in the direction specified (1=+, -1=-)
//Updates position accordingly, and displays actual position according
//to the actual step number.
//

void moveDegrees(long degs)
{
 position+=(degs*direction);		//Bump the current position up or down
 if (position<0) position+=360000;	//Correct underflow in position
 position=(position%360000);		//Correct overflow in position
 increment=(((position*Params.Ratio)+450)/900);
    					//Steps for desired position,
    					//rounded to nearest step

 if (direction==1)			//+ve move
  increment-=stepNumber;		//Calculate size of move
 else					//-ve move
  increment=stepNumber-increment;
    
 if (increment<0)
  increment+=stepsPerRev;		//Correct underflow

 if ((!increment)&&(degs==360000))
  increment=stepsPerRev;		//Special case when Params.Degrees=360000
   
 rampud(increment);
  					//Do the move
 if (stepNumber<0) stepNumber+=stepsPerRev;
  					//Correct underflow
 stepNumber=(stepNumber%stepsPerRev);
  					//Correct overflow
 LCDstepNumber();			//Print actual position in degrees
};

//
//setCurrent()
//Sets current to high (full value of Params.Current) or low (50%)
//

void setCurrent(int percent)
{
 duty=(Params.Current*1024)/100);
 				//Calculate duty cycle bits
 duty=(duty*percent)/100;	//Adjust to percentage of max current
 dutyCycle=(unsigned char)(duty>>2);
 				//Extract top 8 bits to an unsigned char
 CCPR1L=dutyCycle;		//Set the duty cycle
 dutyCycle=(unsigned char)((duty%4)<<4)+0xC;
 				//PWM mode, plus 2 LS bits of duty cycle
 CCP1CON=dutyCycle;		//Set CCP1CON
};

//
// slave()
// Runs slave mode operation
// Sets step, dir, and enable pins to input mode
// Sets current to full if activity is seen
// Sets current to low if no activity seen
//

 void slave()
 {
  int motion;			//Motion indicator
  unsigned int oldbits, newbits;//Last/current value of the step/dir/enable bits
  TRISB=0xFF;			//Set all bits of port B to input

  LCDPrintAt(0,1);		//2nd Line of display
  LCDString("DRIVER MODE ONLY");//Announce driver mode
    
  motion=0;			//Clear motion detector
  oldbits=PORTB&3;		//Get current value of Port B low 3 bits
  setCurrent(Params.Idle);	//Set current to idle
  while(1)			//Do forever
  {
   newbits=PORTB&3;		//Get new bits
   if (newbits != oldbits)	//Any change?
   {
    if (!motion)
    {
     setCurrent(100);		//set current to high
    }
    oldbits=newbits;		//record the change
    motion=32000;		//set motion timer
   }
   if (motion)
   {
    motion -=1;			//Decrement motion timer if non zero
    if (!motion)
    {
     setCurrent(Params.Idle);	//Set idle current if timer has expired
    }
   }
  }
 };
 
 //
 // Continuous()
 // Runs continuous mode operation
 // Ramps up to max speed, holds max speed till "OK" pressed, then ramps back
 // down to min speed.
 // Direction is determined by the <> keys that start the move.
 // Aborts operation if STOP pressed.
 
void continuous()

{
startCont:
 LCDPrintAt(0,1);			//2nd line
 LCDString(ExitString);			//Print exit command
 do
 {
  if (!setSpeeds())			//Wait for meaningful keypress & set speeds
   continue;				//Only make a move if one is wanted
  
  LCDPrintAt(0,1);			//2nd line
  LCDString(ExitString2);		//Lose the <>
  setCurrent(100);			//Motor current to max
 
  rampcont();
 
  waitTime(100000);			//Allow motor to stop
  setCurrent(Params.Idle);		//Set idle current
  PB.B2=0;				//Switch off the driver
  LCDPrintAt(0,1);			//2nd line
  LCDString(ExitString);		//Restore the <>
  newKey=getKey();			//Make sure we know which key was pressed
  goto startCont;			//Loop
 }
 while ((newKey!=Down)&&(newKey!=Up));
 					//Stay in mode till kicked out
};


//
// rampcont function - ramps up/down using the current values for
// start & end speed, holding on the plateau till kicked out by Up or Down.
// direction = 1 for +ve move, -1 for -ve move
//

void rampcont()

{
 long ramp;			//ramp size
  
 ramp=(slow-fast)/rate;		//Find out size of ramp
 ramp=ramp*8;			//Adjust for divide-by-8 in doSteps
 
 delay=slow;			//set initial speed
 
 PD.B0=1;			//Signal "Running"
 dosteps(ramp, rate);
 				//do up ramp

 if (newKey==Abort)
  goto abortRampCont;		//Bomb out if abort pressed

 do
 {
  dosteps(10, 0);		//...and do the plateau
  PA.B3=1;			//Read the up/down keys
 }
 while ((newKey!=Abort) && (PB.B7) && (PB.B6));
  				//...until kicked out

 if (newKey==Abort)
  goto abortRampCont;		//Bomb out if abort pressed

 dosteps(ramp, -rate);
				//do down ramp
abortRampCont:
 PA.B3=0;			//Kill the row drive for up/down testing
 PD.B0=0;			//Signal "Stopped"
};
 
 
//
// Convert Speed procedure
//
// Converts an integer speed value in steps/second
// to an integer value to control the step generation delay loop in doSteps()
//


unsigned int convertSpeed(long speed)

{
 long x=10000000;		//Local scratchpad
 
 x-=(speed*636);		//~63.6 microsecs overhead per main loop
 x=(x/(speed*62)); 		//~6.2 microsecs per delay loop iteration
 return x;
};

//
//Set Speeds procedure
//
//Checks for <> and 1 or 3 keys
//and sets slow, fast speeds and direction variable accordingly. 
//Also exits with a prv or nxt key.
//

char setSpeeds()

{
startSetSpeeds:
 char returnValue=1;			//Assume we will get a "Go" key
 keyChange();
 slow=convertSpeed(Params.Slow);
 					//Assume "normal" speeds
 fast=convertSpeed(Params.Fast);	
 rate=Params.Rate;			//and accelerations

 if (newKey==Right)
  direction=1;				//+ve move, normal
 else if (newKey==Left)
  direction=-1;				//-ve move, normal
 else if (newKey==3)			//+ve move, slow
 {
  direction=1;
  slow=convertSpeed(Params.JogSlow);
  fast=convertSpeed(Params.JogFast);	
  rate=Params.JogRate;
 }
 else if (newKey==1)			//-ve move, slow
 {
  direction=-1;
  slow=convertSpeed(Params.JogSlow);
  fast=convertSpeed(Params.JogFast);	
  rate=Params.JogRate;
 }
 else if ((newKey!=Up)&&(newKey!=Down))
  goto startSetSpeeds;
 else returnValue=0;
 return returnValue;
};

//
//Interrupt handler
//

//const int QuickInt=1;			//Flag the use of quick interrupts

//void Interrupt()
//{
// #asm
// movlw 215
// movwf TMR0				;Preload for 20 microsecs per interrupt
// 					;166 = 40 microsecs
// 					;216 = 20 microsecs
// 					;241 = 10 microsecs
// bcf INTCON, T0IF			; Clear the timer interrupt
// bsf PORTB, 0				; Set the step flag
// nop
// nop
// nop
// nop
// nop
// nop
// nop
// nop
// bcf PORTB, 0				; Clear the step flag
// #asmend
//};
